/****************************************************************************
*  This file is part of the XFLASH device driver application.
*
*  Copyright (c) 2004 by Michael Fischer. All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
****************************************************************************
*
*  Name:    xflash.c
*
*  Remarks: This file system is for an AT45DB041B.
*
*  History:
*
*  23.10.04  mifi   First Version
****************************************************************************/
//#define __XFLASH_C__


#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>

#include <sys/heap.h>
#include <sys/event.h>
#include <sys/thread.h>
#include <sys/atom.h>
#include <sys/device.h>

#include "project.h"
#include "xflash.h"
#include "xpnutfs.h"

/*=========================================================================*/
/*  DEFINE: All Structures and Common Constants                            */
/*=========================================================================*/
/* 
 * SPI memory chip layout, see data sheet. 
 */
 
#define FLASH_PAGE_SIZE	   (PNUT_BLOCK_SIZE / 2)
  
#define MAX_PAGE_COUNT      2048

/* 
 * SPI memory chip select, see schematics. 
 */
#define SPIMEM_CS_PORT      PORTB
#define SPIMEM_CS_DDR       DDRB
#define SPIMEM_CS           0x10


#define SPI_CS_LOW()        PORTB |= SPIMEM_CS
#define SPI_CS_HIGH()       PORTB &= ~SPIMEM_CS


/*=========================================================================*/
/*  DEFINE: Definition of all local Data                                   */
/*=========================================================================*/
static HANDLE hXFLASHSemaphore;

/*=========================================================================*/
/*  DEFINE: Definition of all local Procedures                             */
/*=========================================================================*/
/***************************************************************************/
/*  XFLASHLock                                                             */
/***************************************************************************/
static void XFLASHLock (void)
{
  NutEventWait(&hXFLASHSemaphore, 0);
} /* XFLASHLock */

/***************************************************************************/
/*  XFLASHFree                                                             */
/***************************************************************************/
static void XFLASHFree (void)
{
  NutEventPost(&hXFLASHSemaphore);
} /* XFLASHFree */

/***************************************************************************/
/*  FATSemaInit                                                            */
/***************************************************************************/
static void XFLASHSemaInit (void)
{
  NutEventPost(&hXFLASHSemaphore);
} /* XFLASHSemaInit */

/***************************************************************************/
/*  SpiByte                                                                */
/*                                                                         */
/*  Send a data byte over the SPI bus,                                     */
/*  at the same time a byte is read back                                   */
/***************************************************************************/ 
static BYTE SpiByte (BYTE bData)
{
  /*
   * Put the byte in the transmit register
   */ 
  SPDR = bData;                 
  
  /*
   * Wait for serial transfer is complete
   */
  while ((inb(SPSR) & 0x80) == 0);

  
  /*
   * Read the byte from the receive register
   */
  return (SPDR);               
} /* SpiByte */ 

/***************************************************************************/
/*  ReadStatus                                                             */
/*                                                                         */
/*  Read the status from the flash memory.                                 */
/***************************************************************************/
BYTE SpiMemStatus(void)
{
  BYTE bData;

  SPI_CS_LOW();
  
  SpiByte(0xD7);
  bData = SpiByte(0);
  
  SPI_CS_HIGH();
  
  return (bData);
} /* SpiMemStatus */

/***************************************************************************/
/*  XflashInit                                                             */
/***************************************************************************/
int XflashInit(void)
{
  int           nError = XFLASH_ERROR;
  volatile BYTE bValue;
    
  XFLASHSemaInit();

  /*
   * Init SPI memory chip select
   */     
  SPIMEM_CS_PORT &= ~SPIMEM_CS;
  SPIMEM_CS_DDR  |=  SPIMEM_CS;
     
  /* 
   * Init SS pin. When configured as input, we will lose master
   * mode, if this pin goes low. Thus we switch on the pull-up.
   */
  if(bit_is_clear(DDRB, 0)) {
    sbi(PORTB, 0);
  }

  /* Set SCK output. */
  sbi(DDRB, 1);
  /* Set MOSI output. */
  sbi(DDRB, 2);
  /* Enable MISO pullup. */
  sbi(PORTB, 3);

  /* Set double speed. */
  SPCR = (1 << SPI2X);

  SPCR = (1 << SPE) | (1 << MSTR);  /* enable SPI as master, set clk divider */
                                    /* set to max speed */
  
  /* Clean-up the status. */
  outb(SPSR, inb(SPSR));
  bValue = inb(SPDR);
    
  /* Read the status register for a rudimentary check. */
  /* Check if this is a AT45DB041B                     */
  bValue = SpiMemStatus();
  if ((bValue & 0x80) == 0x80) {
    bValue = (bValue >> 2) & 0x0F;
    if(bValue == 7) {
      nError = XFLASH_OK;
    }
  }

  return (nError); 
} /* XflashInit */

/***************************************************************************/
/*  WriteBlock                                                             */
/*                                                                         */
/*  Write a block with the size wSize to the spi bus                       */
/***************************************************************************/
void WriteBlock (BYTE *pBuffer, unsigned wSize)
{
  unsigned i=0;
  BYTE *Ptr = pBuffer;
   
  while(wSize--) {
    outb(SPDR, Ptr[i]);
    i++;
    while ((inb(SPSR) & 0x80) == 0);
  }
} /* WriteBlock */

/***************************************************************************/
/*  ReadBlock                                                              */
/*                                                                         */
/*  Read a block with the size wSize from the spi bus.                     */
/***************************************************************************/
void ReadBlock (BYTE *pBuffer, unsigned wSize)
{
  unsigned i=0;
  BYTE *Ptr = pBuffer;
  
  while(wSize--) {
    outb(SPDR, 0x00);
    while ((inb(SPSR) & 0x80) == 0);
    Ptr[i] = inb(SPDR);
    i++;
  }
} /* ReadBlock */

/***************************************************************************/
/*  PageWrite                                                              */
/*                                                                         */
/*  Write a block of data with the size of wSize to the page wPage.        */
/***************************************************************************/
int PageWrite (unsigned wPage, BYTE *pBuffer, unsigned wSize, unsigned wOffset)
{
  BYTE bByte1;
  BYTE bByte2;
  BYTE bByte3;
  
//  printf("Xflash wPage: %u\n",wPage);
  
  bByte1 = LOBYTE(wPage >> 7);
  
  bByte2 = LOBYTE(wPage << 1);  
  if (wOffset & 0x100) {
    bByte2 |= 0x01;
  }
  
  bByte3 = LOBYTE(wOffset);
  
  SPI_CS_LOW();

  /* Select mode */
  SpiByte(0x82);
  SpiByte(bByte1);
  SpiByte(bByte2);
  SpiByte(bByte3);
  
  /* Write data */
  WriteBlock(pBuffer, wSize);

  SPI_CS_HIGH();  

  while((SpiMemStatus() & 0x80) == 0);

  return (XFLASH_OK);
} /* PageWrite */


/***************************************************************************/
/*  PageRead                                                               */
/*                                                                         */
/*  Read a block of data with the size of wSize from the page wPage.       */
/***************************************************************************/
int PageRead (unsigned wPage, BYTE *pBuffer, unsigned wSize, unsigned wOffset)
{
  BYTE bByte1;
  BYTE bByte2;
  BYTE bByte3;    
  
 
//  printf("Xflash rPage: %u\n",wPage);
  bByte1 = LOBYTE(wPage >> 7);
  
  bByte2 = LOBYTE(wPage << 1);  
  if (wOffset & 0x100) {
    bByte2 |= 0x01;
  }
  
  bByte3 = LOBYTE(wOffset);

  while((SpiMemStatus() & 0x80) == 0);

  SPI_CS_LOW();  

  SpiByte(0xE8);
  SpiByte(bByte1);
  SpiByte(bByte2);
  SpiByte(bByte3);
  SpiByte(0x00);
  SpiByte(0x00);
  SpiByte(0x00);
  SpiByte(0x00);
  ReadBlock(pBuffer, wSize);

  SPI_CS_HIGH();  

  return (XFLASH_OK);
} /* PageRead */

int PageReadWrite (unsigned wPage, BYTE *pBuffer, unsigned wSize, unsigned wOffset)
{
  BYTE *buff;
  unsigned i;
  
  if ((buff = (u_char *)NutHeapAlloc(FLASH_PAGE_SIZE)) == NULL) {
	 printf("Error heap XFlash\n");
	 return -1;
  }  
  PageRead (wPage, buff, FLASH_PAGE_SIZE, 0);
  for (i=0; i<wSize; i++) buff[wOffset + i] = pBuffer[i];
  PageWrite (wPage, buff, FLASH_PAGE_SIZE, 0);

  NutHeapFree(buff);
  
  return (XFLASH_OK);
}

/***************************************************************************/
/*  PageWrite2                                                              */
/*                                                                         */
/*  Write a block of data with the size of wSize to the page wPage.        */
/***************************************************************************/
int PageWrite2 (unsigned wPage, BYTE *pBuffer, unsigned wSize, unsigned wOffset)
{
  unsigned endP;
  
  wPage <<= 1; 			  			 // double nr pages for one one block
  endP = wSize + wOffset; 			 // end point of writing
  if (wOffset < FLASH_PAGE_SIZE) {   // check if to start in page 1
    if (endP <= FLASH_PAGE_SIZE) {	 // check if only page 1
	  PageWrite (wPage++,pBuffer,wSize,wOffset);	// write first page
	}  
	else {		// partly page 1, partly page 2
	  PageWrite (wPage,pBuffer,FLASH_PAGE_SIZE-wOffset,wOffset);
	  wPage++;
	  PageWrite (wPage,pBuffer+FLASH_PAGE_SIZE-wOffset,endP - FLASH_PAGE_SIZE,0);  
	}
  }
  else { 		// Only page 2
    wPage++;	   		
    PageWrite (wPage,pBuffer+FLASH_PAGE_SIZE-wOffset,wSize,wOffset-FLASH_PAGE_SIZE);
  }	
  return (XFLASH_OK);
} /* PageWrite2 */

int PageRead2 (unsigned wPage, BYTE *pBuffer, unsigned wSize, unsigned wOffset)
{
  // see PageWrite2()
  unsigned endP;
  
  wPage <<= 1; 			
  endP = wSize + wOffset; 
  if (wOffset < FLASH_PAGE_SIZE) { 
    if (endP <= FLASH_PAGE_SIZE) {	 
	  PageRead (wPage++,pBuffer,wSize,wOffset);	
	}  
	else {
	  PageRead (wPage,pBuffer,FLASH_PAGE_SIZE-wOffset,wOffset);
	  wPage++;		
	  PageRead (wPage,pBuffer+FLASH_PAGE_SIZE-wOffset,endP - FLASH_PAGE_SIZE,0);  
	}
  }
  else { 
    wPage++;
    PageRead (wPage,pBuffer+FLASH_PAGE_SIZE-wOffset,wSize,wOffset-FLASH_PAGE_SIZE);
  }	
  return (XFLASH_OK);
} /* PageRead2 */

int PageReadWrite2 (unsigned wPage, BYTE *pBuffer, unsigned wSize, unsigned wOffset)
{
  BYTE *buff;
  unsigned i;
  
  if ((buff = (u_char *)NutHeapAlloc(PNUT_BLOCK_SIZE)) == NULL) {
	 printf("Error heap XFlash\n");
	 return -1;
  }  
  PageRead2 (wPage, buff, PNUT_BLOCK_SIZE, 0);
  for (i=0; i<wSize; i++) buff[wOffset + i] = pBuffer[i];
  PageWrite2 (wPage, buff, PNUT_BLOCK_SIZE, 0);

  NutHeapFree(buff);
  
  return (XFLASH_OK);
}


